package Impedance::W;
#$Id: W.pm,v 1.19 2012/10/12 10:54:34 sofc Exp $
#####################################################################
#
# Package Impedance::W;
#
# This module implements Impedance::Base for diffusion type elements
# (W, O, G and T).
# W is the Warburg element, O is the Finite length Warburg element,
# G is the Gerisher element and T is the 'Bounded Warburg' element.
#
# To obtain an instance call one of the constructors:
#
# $id = Impedance::W->new($nr);
#
# $id = Impedance::O->new($nr);
#
# $id = Impedance::G->new($nr);
#
# $id = Impedance::T->new($nr);
#
# The module denines no additional member functions.
#
#####################################################################
use strict;
use warnings;
use Impedance::Base;
use Math::Complex;
our @ISA = qw(Impedance::Base);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'W';
    $self->{'_tags'} = ['Y'];
    $self->{'_desc'}{'name'} = "Warburg diffusion element (infinite linear diffusion)";
    $self->{'_desc'}{'Y'} = "Admittance in S*s^(1/2)";
    $self->{'_f_imax'} = 1e-100;
    bless $self,$class;
    return $self;
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("Y_$n = ".$self->value('Y'),"fy_$n = 1.0");
    push @l,'z_'.$n.'(x)={1,-1} / (2**0.5*Y_'.$n.'*fy_'.$n.'*(pi*2*x)**0.5)';
    return join("\n",@l)."\n";
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return cplx(1,-1) / (2**0.5 * $self->value('Y') * (2 * pi() * $f)**0.5);
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fy_$n");
}

package Impedance::O;
use Math::Complex;
### Finite length warburg impedance ###
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'O';
    $self->{'_tags'} = ['Y','B'];
    $self->{'_helptags'} = ['B','Y'];
    $self->{'_desc'}{'name'} = "Warburg diffusion element of finite length";
    $self->{'_desc'}{'Y'} = "'Admittance' of element in S";
    $self->{'_desc'}{'B'} = "'Time constant' in s^(1/2)";
    bless $self,$class;
    return $self;
}

sub f_imax()
{
    my $self = shift;
    return 0.4 / ($self->value('B') ** 2);
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("Y_$n = ".$self->value('Y'),"fy_$n = 1.0","B_$n = ".$self->value('B'),"fb_$n = 1.0");
    push @l,'tnh(x) = (abs(x)>100 ? {1,0} : tanh(x))';
    push @l,'z_'.$n.'(x)=tnh(B_'.$n.' * fb_'.$n.' * ({0,1}*2*pi*x)**0.5 ) / (Y_'.$n.' * fy_'.$n.' * ({0,1}*2*pi*x)**0.5)';
    return join("\n",@l)."\n";
}


sub save()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ();
    push @l,"Final_B_$n = B_$n * fb_$n";
    push @l,"Final_Y_$n = Y_$n * fy_$n";
    push @l,"Final_Fmax_$n = 0.4 / (B_$n * B_$n)";
    return join("\n",@l)."\n";
}

sub my_tanh($)
{
    my $self = shift;
    my $arg = shift;
    if ($arg > 100)
    {
	return cplx(1,0);
    }
    else
    {
	return tanh($arg);
    }
}


sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return $self->my_tanh($self->value('B') * (cplx(0,1) * 2 * pi() * $f)**0.5) / ($self->value('Y') * (cplx(0,1) * 2 * pi() * $f)**0.5);
}

sub B_start($$$)
{
    ### Returns a guess at the 'B' parameter of the elemenet ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $x = $l[0];  ### Summit frequency in Hz   ##
    return sqrt(0.4/$x);
}

sub Y_start($$$)
{
    ### Returns a guess at the 'capacitance' of the element ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $B = sqrt(0.4/$l[0]); ### gest guess at B parameter                   ##
    my @m = split(/[\s\t]+/,$maxf);
    @l = split(/[\s\t]+/,$minf);
    my $z_max = $l[1] - $m[1];  ### max impedance (intercept with x for element ##
    return $B/$z_max;
}

sub helpfunction
{
    my $self = shift;
    unless (@_)
    {
	return $self->helperfunctions();
    }
    my $tag = shift;
    return "Select minimum frequency, sumit frequency and maximum frequency for arc" unless (@_);
    return $self->Y_start(@_) if ($tag eq 'Y');
    return $self->B_start(@_) if ($tag eq 'B');
    return undef;
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fy_$n","fb_$n");
}

package Impedance::T;
use Math::Complex;
### Finite capacity warburg impedance ###
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'T';
    $self->{'_tags'} = ['Y','B'];
    $self->{'_helptags'} = ['Y','B'];
    $self->{'_desc'}{'name'} = "Warburg diffusion element of finite capacity (bounded warburg)";
    $self->{'_desc'}{'Y'} = "'Admittance' of element in S";
    $self->{'_desc'}{'B'} = "'Time constant' in s^(1/2)";
    $self->{'_B_hash'} = {
	     95 => 0.1,
	     25 => 0.2,
	     10 => 0.3,
	     5 => 0.4,
	     3 => 0.5,
	     2 => 0.6,
	     1.5 => 0.7,
	     1.25 => 0.8,
	     1 => 0.9,
	     0.75 => 0.9,
	     0.6 => 1.1,
	     0.5 => 1.2,
	     0.4 => 1.3,
	     0.3 => 1.5,
	     0.15 => 2,
	     0.1 => 3,
	     0.05 => 5,
	     0.01 => 10,
	     0.002 => 20,
	     };
    bless $self,$class;
    return $self;
}

sub f_imax()
{
    my $self = shift;
    foreach my $f (reverse sort {$a <=> $b} keys %{$self->{'_B_hash'}})
    {
	next if $self->{'_B_hash'}{$f} < $self->value('B');
	return $f;
    }
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("Y_$n = ".$self->value('Y'),"fy_$n = 1.0","B_$n = ".$self->value('B'),"fb_$n = 1.0");
    push @l,'cth(x) = (abs(x)>50 ? {1,0} : cosh(x)/sinh(x))';
    push @l,'z_'.$n.'(x)=cth(B_'.$n.' * fb_'.$n.' * ({0,1}*2*pi*x)**0.5 ) / (Y_'.$n.' * fy_'.$n.' * ({0,1}*2*pi*x)**0.5)';
    return join("\n",@l)."\n";
}

sub my_coth($)
{
    my $self = shift;
    my $val = shift;
    if ($val > 50)
    {
	return cplx(1,0);
    }
    else
    {
	return cosh($val)/sinh($val);
    }
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return $self->my_coth($self->value('B') * (cplx(0,1) * 2 * pi() * $f)**0.5) / ($self->value('Y') * (cplx(0,1) * 2 * pi() * $f)**0.5);
}

sub B_start($$$)
{
    ### Returns a guess at the 'B' parameter of the elemenet ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $x = $l[0];  ### 'kink' frequency in Hz   ##
    foreach my $f (reverse sort {$a <=> $b} keys %{$self->{'_B_hash'}})
    {
	next if $f > $x;
	return $self->{'_B_hash'}{$f};
    }
}

sub Y_start($$$)
{
    ### Returns a guess at the 'capacitance' of the element ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $B = $self->B_start($minf,$sumf,$maxf);
    my @m = split(/[\s\t]+/,$maxf);
    @l = split(/[\s\t]+/,$minf);
    my $z_max = $l[1] - $m[1];  ### max impedance (intercept with x for element ##
    return $B/(3 * $z_max);
}

sub helpfunction
{
    my $self = shift;
    unless (@_)
    {
	return $self->helperfunctions();
    }
    my $tag = shift;
    return "Select minimum frequency, bend frequency and maximum frequency" unless (@_);
    return $self->Y_start(@_) if ($tag eq 'Y');
    return $self->B_start(@_) if ($tag eq 'B');
    return undef;
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fy_$n","fb_$n");
}


package Impedance::G;
use Math::Complex;
### Gerisher impedance ###
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'G';
    $self->{'_tags'} = ['Y','k'];
    $self->{'_helptags'} = ['Y','k'];
    $self->{'_desc'}{'name'} = "Gerisher diffusion element";
    $self->{'_desc'}{'Y'} = "'Admittance' of element in S*s^(1/2)";
    $self->{'_desc'}{'k'} = "'Rate constant' in s^(-1";
    bless $self,$class;
    return $self;
}

sub f_imax()
{
    my $self = shift;
    return $self->value('k')/4;
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("Y_$n = ".$self->value('Y'),"fy_$n = 1.0","k_$n = ".$self->value('k'),"fk_$n = 1.0");
    push @l,'z_'.$n.'(x)=1.0 / ( Y_'.$n.' * fy_'.$n.' * (k_'.$n.' * fk_'.$n.' + {0,1}*2*pi*x)**0.5)';
    return join("\n",@l)."\n";
}

sub save()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ();
    push @l,"Final_k_$n = k_$n * fk_$n";
    push @l,"Final_Y_$n = Y_$n * fy_$n";
    push @l,"Final_Fmax_$n = Final_k_$n / 4";
    return join("\n",@l)."\n";
}


sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return cplx(1,0)/($self->value('Y') * ($self->value('k') + cplx(0,1) * 2 * pi() * $f)**0.5);
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fy_$n","fk_$n");
}

sub k_start($$$)
{
    ### Returns a guess at the 'B' parameter of the elemenet ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $x = $l[0];  ### Summit frequency in Hz   ##
    return 4 * $x;
}

sub Y_start($$$)
{
    ### Returns a guess at the 'capacitance' of the element ###
    my $self = shift;
    my $minf = shift;
    my $sumf = shift;
    my $maxf = shift;
    my @l = split(/[\s\t]+/,$sumf);
    my $k = 4 * $l[0]; ### gest guess at k parameter                   ##
    @l = split(/[\s\t]+/,$minf);
    my @m = split(/[\s\t]+/,$maxf);
    my $z_max = $l[1] - $m[1];  ### max impedance (intercept with x for element ##
    return 1 / ( $z_max * $k**0.5 );
}

sub helpfunction
{
    my $self = shift;
    unless (@_)
    {
	return $self->helperfunctions();
    }
    my $tag = shift;
    return "Select minimum frequency, sumit frequency and maximum frequency for arc" unless (@_);
    return $self->Y_start(@_) if ($tag eq 'Y');
    return $self->k_start(@_) if ($tag eq 'k');
    return undef;
}

return 1;

